/*
 * args.c
 */

static char *rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/args.c,v 1.1 1994/11/01 03:50:35 tuecke Exp $";

#include "internal.h"


/*
 * remove_arguments()
 *
 * Remove 'args_to_remove' argument from argc/argv, starting
 * at 'current_arg_index'.
 */
static void remove_arguments(int *argc,
			     char **argv[],
			     int current_arg_index,
			     int args_to_remove)
{
    int i;
    for (i = 0; i < args_to_remove; i++)
    {
	(*argv)[current_arg_index + i]
	    = (*argv)[current_arg_index + args_to_remove + i];
    }
    *argc -= args_to_remove;
} /* remove_arguments() */


/*
 * process_argument()
 *
 * Process the ports0 argument, 'ports0_arg', which is at location
 * 'current_arg_index' in argc/argv.
 *
 * Return the number of arguments which are recognized, starting at
 * current_arg_index, which can be removed.
 */
static int process_argument(char *ports0_arg,
			    int current_arg_index,
			    int argc,
			    char *argv[])
{
    int args_removed;
    
    /* Give the thread module a shot at the arguments */
    if ((args_removed = _p0_thread_process_argument(ports0_arg,
						    current_arg_index,
						    argc,
						    argv)) > 0)
    {
	return (args_removed);
    }

    /* Try any other modules here... */
       
    return (0);
} /* process_argument() */


/*
 * _p0_parse_args
 *
 * Parses ports0 arguments and removes them from the argc/argv.
 * Ports arguments can take on one of 2 forms.  Given that the
 * package_id="foo":
 *      1.  -foo_argument : this can exist anywhere on the command line.
 *      2.  -foo -argument ... ... -foo_end : any arguments between
 *          -foo and -foo_end are package arguments.
 */
void _p0_parse_args(int  *argc,
		    char **argv[],
		    char *package_id_arg)
{
    char package_id[64];
    char package_id_end[64];
    char package_id_prefix[64];
    int  package_id_len;
    int  package_id_prefix_len;
    char *current_arg;
    int current_arg_index;
    int args_removed;
    char *ports0_arg;
    int package_id_start_index;

    _p0_thread_process_arguments_init();

    sprintf(package_id, "-%s", package_id_arg);
    sprintf(package_id_end, "%s_end", package_id);
    sprintf(package_id_prefix, "%s_", package_id);
    package_id_len = strlen(package_id);
    package_id_prefix_len = strlen(package_id_prefix);

    package_id_start_index = -1;
    current_arg_index = 1;

    while (current_arg_index < *argc)
    {	
	current_arg = (*argv)[current_arg_index];
	
        if (strcmp(current_arg, package_id) == 0)
        {
	    /*
	     * Got package_id start.
	     * All subsequent arguments until package_id_end
	     * are ports0 arguments.
	     */
            package_id_start_index = current_arg_index;
	    current_arg_index++;
	}
        else if (strcmp(current_arg, package_id_end) == 0)
	{
	    /* Got package_id end. */
	    if (package_id_start_index == -1)
	    {
		ports0_fatal("Error: Got %s argument without %s argument.",
			     package_id_end, package_id);
		exit(1);
	    }
	    if ((package_id_start_index + 1) == current_arg_index)
	    {
		/*
		 * package_id start and end are right next to each
		 * other, so remove them from the argument list.
		 */
		remove_arguments(argc, argv, package_id_start_index, 2);
		current_arg_index = package_id_start_index;
	    }
	    else
	    {
		/*
		 * There are still arguments between package_id start and end,
		 * so leave them and go on with next argument.
		 */
		current_arg_index++;
	    }
	    package_id_start_index = -1;
	}
	else
	{
	    /* This is not the package_id start or end marker */
	    ports0_arg = (char *) NULL;
	    if (strncmp(current_arg, package_id_prefix,
			package_id_prefix_len) == 0)
	    {
		/*
		 * This is a ports0 argument because it starts with
		 * the package_id_prefix.
		 */
		ports0_arg = current_arg + package_id_prefix_len;
		if (ports0_arg[0] == '\0')
		{
		    ports0_fatal("Error: Illegal argument: %s\n", current_arg);
		}
	    }
	    else if (package_id_start_index != -1)
	    {
		/*
		 * We are between package_id start and end.
		 * So this is a ports0 argument if it starts with a "-".
		 */
		if (current_arg[0] == '-')
		{
		    /* Starts with "-", so its a ports0 argument */
		    ports0_arg = current_arg + 1;
		}
		else
		{
		    /* Go on to the next argument */
		    current_arg++;
		}
	    }
	    else
	    {
		/*
		 * This is not a ports0 argument.
		 * So just skip it an go on to the next argument.
		 */
		current_arg_index++;
	    }

	    if (ports0_arg != (char *) NULL)
	    {
		/*
		 * I've got a ports0 argument.
		 * See if I know what to do with it.
		 * If so, then remove it from argv.
		 * If not, then just leave it and go on with next argument.
		 */
		args_removed = process_argument(ports0_arg,
						current_arg_index,
						*argc,
						*argv);
		if (args_removed > 0)
		{
		    remove_arguments(argc, argv,
				     current_arg_index, args_removed);
		    /* Leave current_arg_index the same */
		}
		else
		{
		    current_arg_index++;
		}
	    }
	}
    }

    /*
     * If the package_id start is at the end of the command line,
     * this implies that there was no package_id end (it was implicit),
     * and that all arguments after the package_id start were removed.
     * So remove the package_id start from argc/argv.
     */
    if (   (package_id_start_index != -1)
	&& (package_id_start_index == *argc-1) )
    {
	remove_arguments(argc, argv, package_id_start_index, 1);
    }

} /* _p0_parse_args */


/*
 * ports0_usage_message()
 *
 * Print a usage message for each ports0 argument.
 *
 * This will loop through the various modules, allowing them to
 * print a usage message also.
 */
void ports0_usage_message()
{
    _p0_thread_usage_message();
} /* ports0_usage_message() */


/*
 * ports0_new_process_params()
 *
 * Add any ports0 arguments to 'buf' that should be passed to new
 * ports0 programs.  'buf' contains 'size' bytes.
 *
 * Return: The number of characters written to 'buf'.
 */
int ports0_new_process_params(char *buf, int size)
{
    int n_left;
    int n_added;
    char *b;
    char tmp_buf1[256];
#if defined(PORTS0_DEBUG) || defined(PORTS0_MALLOC_DEBUG)    
    char tmp_buf2[256];
#endif    

    n_left = size;
    b = buf;

    /*
     * Put the basic ports0 arguments into the buffer
     */
    tmp_buf1[0] = '\0';

#ifdef PORTS0_DEBUG
    if (_p0_debug_level >= 0)
    {
	sprintf(tmp_buf2, "-Dports0 %d ", _p0_debug_level);
	strcat(tmp_buf1, tmp_buf2);
    }
#endif
    
#ifdef PORTS0_MALLOC_DEBUG    
    if (_p0_malloc_debug_level >= 0)
    {
	sprintf(tmp_buf2, "-Dmalloc %d ", _p0_malloc_debug_level);
	strcat(tmp_buf1, tmp_buf2);
    }
#endif

    n_added = strlen(tmp_buf1);
    if (n_added > n_left)
    {
	ports0_fatal("ports0_new_process_params(): Internal error: Not enough room in buffer for arguments\n");
    }

    strcpy(b, tmp_buf1);
    b += n_added;
    n_left -= n_added;

    
    /*
     * Put arguments for the various modules.
     */
    n_added = _p0_thread_new_process_params(b, n_left);
    b += n_added;
    n_left -= n_added;
    
    return (size - n_left);
} /* ports0_new_process_params() */
