/*
 * th_p.c
 *
 * General purpose pthreads module.
 */

static char *rcsid = "$Header: /Net/bioserv.compbio.caltech.edu/bcomp/src/master/nexus/ports0/th_p.c,v 1.1 1994/11/01 03:53:16 tuecke Exp $";

#include "internal.h"

#define MAX_ERR_SIZE 80
#define _P0_THREAD_GRAN 256
#define _P0_USER_THREAD 0

#if PORTS0_DEBUG
static int hasThreads = 0;
#endif

#ifdef PORTS0_ARCH_PARAGON
/* Paragon doesn't support changing the scheduler or set priority */
#define SCHED_RR   0
#define SCHED_FIFO 1
#define PRI_FIFO_MIN 0
#define PRI_FIFO_MAX 0
#define PRI_RR_MIN 0
#define PRI_RR_MAX 0
#define MUTEX_NONRECURSIVE_NP 0
#define PTHREAD_DEFAULT_SCHED 0
#endif /* PORTS0_ARCH_PARAGON */

static bool_t		preemptive_threads;
static bool_t		arg_got_sched;
static int	 	scheduler;
static int		priority_min;
static int		priority_max;
static int		priority_mid;
static pthread_attr_t   default_thread_attr;
#ifdef PORTS0_HACK_STACK_SIZE
static long		arg_stack_size;
static long		stack_size;
#endif

static bool_t		using_handler_thread = PORTS0_FALSE;
static pthread_t	handler_thread;
static bool_t		handler_thread_done = PORTS0_FALSE;
static void *		idle_thread_func(void *arg);

/*
 * Free list of _p0_thread_t structures
 */
static _p0_thread_t *	thread_freelist;
static ports0_mutex_t	thread_mem_mutex;
static int              next_thread_id;

static void *		thread_starter(void *temparg);
static _p0_thread_t *	new_thread(void);
static void		set_tsd(_p0_thread_t *);

static void (*idle_func_save)();

#define _P0_THR_SUCCESS 0
#define _p0_test_rc(rc, msg) \
do {\
    if (rc != _P0_THR_SUCCESS) _p0_report_bad_rc(rc, msg); \
} while(0)

/*
 * _p0_thread_process_arguments_init()
 *
 * Initialize any variables that will be used to hold
 * my configuration information extracted from arguments.
 */
void _p0_thread_process_arguments_init()
{
    arg_got_sched = PORTS0_FALSE;
    scheduler = SCHED_RR;
#ifdef PORTS0_HACK_STACK_SIZE
    arg_stack_size = -1;
#endif    
} /* _p0_thread_process_arguments_init() */


/*
 * _p0_thread_process_argument()
 *
 * If ports0_package_getarg(current_arg) is destined for my module,
 * then extract it into my own variables.
 *
 * Return:	A new current argument.
 */
int _p0_thread_process_argument(char *arg_base,
				int current_arg,
				int argc,
				char *argv[])
{
    char *arg;
    
    if ((strcmp(arg_base, "sched") == 0) &&
	     ((current_arg + 1) < argc))
    {
	arg = argv[current_arg + 1];
	if (strcmp(arg, "fifo") == 0)
	{
	    scheduler = SCHED_FIFO;
	}
	else if (strcmp(arg, "rr") == 0)
	{
	    scheduler = SCHED_RR;
	}
	else
	{
	    ports0_fatal("-sched argument must use either \"fifo\" or \"rr\"\n");
	}
	arg_got_sched = PORTS0_TRUE;
	current_arg += 2;
    }
#ifdef PORTS0_HACK_STACK_SIZE
    else if ((strcmp(arg_base, "stack") == 0) &&
	     ((current_arg + 1) < argc))
    {
	arg_stack_size = atol(argv[current_arg + 1]);
	current_arg += 2;
    }
#endif
    return (current_arg);
} /* _p0_thread_process_arguments() */


/*
 * _p0_thread_usage_message()
 *
 * Print a usage message for this module's arguments to stdout.
 */
void _p0_thread_usage_message()
{
    printf("    -sched <string>           : Specify which scheduler to use:\n");
    printf("                                <string> = rr: round robin (default)\n");
    printf("                                <string> = fifo: first in first out\n");
#ifdef PORTS0_HACK_STACK_SIZE
    printf("    -stack <integer>          : Set the thread stack size.\n");
#endif    
} /* _p0_thread_usage_message() */


/*
 * _p0_thread_new_process_params()
 *
 * Write any new process parameters for this module into 'buf',
 * up to a maximum of 'size' characters.
 *
 * Return:	The number of characters written into 'buf'.
 */
int _p0_thread_new_process_params(char *buf, int size)
{
    char tmp_buf1[256];
    char tmp_buf2[128];
    int n_added = 0;
    
    tmp_buf1[0] = '\0';
    
    if (arg_got_sched)
    {
	if (scheduler == SCHED_FIFO)
	{
	    strcat(tmp_buf1, "-sched fifo ");
	}
	else /* (scheduler == SCHED_RR) */
	{
	    strcat(tmp_buf1, "-sched rr ");
	}
    }

#ifdef PORTS0_HACK_STACK_SIZE
    if (arg_stack_size > 0)
    {
	sprintf(tmp_buf2, "-stack %ld ", arg_stack_size);
	strcat(tmp_buf1, tmp_buf2);
    }
#endif
    
    n_added = strlen(tmp_buf1);
    if (n_added > size)
    {
	ports0_fatal("_p0_thread_new_process_params(): Internal error: Not enough room in buffer for arguments\n");
    }
    
    strcpy(buf, tmp_buf1);
    
    return (n_added);
    
} /* _p0_thread_new_process_params() */


/*
 * _p0_thread_preinit()
 *
 * If you need to call a thread package initialization routine, then
 * do it here.  This is called as the very first thing in ports0_init().
 *
 * Note: You should not use ports0 thread stuff until after
 * _p0_thread_init() is called.
 */
void _p0_thread_preinit( void )
{
#ifdef PORTS0_CALL_PTHREAD_INIT
    /*
     * FSU pthreads uses its own read/write/malloc routines, which
     * might get called during the argument handling, before
     * _p0_thread_init() is called.
     */
    pthread_init();
#endif /* PORTS0_CALL_PTHREAD_INIT */
} /* _p0_thread_preinit() */


/*
 * _p0_thread_init()
 *
 * This should be used to initialize all the thread related things
 * that Ports0 will be using, including things that might be
 * specified by arguments.
 */
void _p0_thread_init( void )
{
    int rc;
    _p0_thread_t *thread;

#ifdef PORTS0_HACK_STACK_SIZE
    if (arg_stack_size > 0)
    {
	stack_size = arg_stack_size;
    }
    else
    {
	stack_size = (long) PORTS0_HACK_STACK_SIZE;
    }
#endif
    
#ifdef PORTS0_ARCH_DCE_THREADS
    /*
     * Set the default attributes for mutex and condition
     */
    rc = pthread_mutexattr_create(&(ports0_all_global_vars.mutexattr));
    _p0_test_rc(rc, "PORTS0: pthread_mutexattr_create() failed\n");

#ifndef PORTS0_ARCH_PARAGON
    rc = pthread_mutexattr_setkind_np(&(ports0_all_global_vars.mutexattr),
				      MUTEX_NONRECURSIVE_NP );
    _p0_test_rc(rc, "PORTS0: pthread_mutexattr_setkind() failed\n");
#endif /* PORTS0_ARCH__PARAGON */
    
    ports0_all_global_vars.condattr = pthread_condattr_default;
#endif /* PORTS0_ARCH_DCE_THREADS */

#ifndef PORTS0_THREAD_SAFE_STDIO
    ports0_mutex_init(&ports0_reentrant_mutex, (ports0_mutexattr_t *) NULL);
#endif /* PORTS0_THREAD_SAFE_STDIO */

    /*
     * Define the parameters for the scheduler being used
     */
    if (scheduler == SCHED_FIFO)
    {
        preemptive_threads = PORTS0_FALSE;
#ifdef PORTS0_ARCH_DCE_THREADS	
	priority_min = PRI_FIFO_MIN;
	priority_max = PRI_FIFO_MAX;
#else
        priority_min = sched_get_priority_min( SCHED_FIFO );
	priority_max = sched_get_priority_max( SCHED_FIFO );
#endif /* PORTS0_ARCH_DCE_THREADS */
	priority_mid = (priority_min + priority_max) / 2;
    }
    else /* (scheduler == SCHED_RR) */
    {
	preemptive_threads = PORTS0_TRUE;
#ifdef PORTS0_ARCH_DCE_THREADS
	priority_min = PRI_RR_MIN;
	priority_max = PRI_RR_MAX;
#else
        priority_min = sched_get_priority_min( SCHED_RR );
	priority_max = sched_get_priority_max( SCHED_RR );
#endif /* PORTS0_ARCH_DCE_THREADS */
	priority_mid = (priority_min + priority_max) / 2;
    }

    /*
     * Setup the default thread attributes
     */
#ifdef PORTS0_ARCH_DCE_THREADS
    rc = pthread_attr_create(&default_thread_attr);
    _p0_test_rc( rc, "PORTS0: pthread_attr_create() failed\n" );
#else
    rc = pthread_attr_init(&default_thread_attr);
    _p0_test_rc( rc, "PORTS0: pthread_attr_init() failed\n" );
#endif /* PORTS0_ARCH_DCE_THREADS */

#ifndef PORTS0_ARCH_PARAGON
    rc = pthread_attr_setinheritsched(&default_thread_attr,
				      PTHREAD_DEFAULT_SCHED);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setinheritsched() failed\n" );
    rc = pthread_attr_setsched(&default_thread_attr, scheduler);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setsched() failed\n" );
    rc = pthread_attr_setprio(&default_thread_attr, priority_mid);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setprio() failed\n" );
#endif /* PORTS0_ARCH_PARAGON */
#ifdef PORTS0_HACK_STACK_SIZE
    rc = pthread_attr_setstacksize(&default_thread_attr, stack_size);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setstacksize() failed\n" );
#endif

    /*
     * Make this initial thread have the default thread attributes
     */
#ifdef PORTS0_ARCH_DCE_THREADS    
#ifndef PORTS0_ARCH_PARAGON
    pthread_setscheduler(pthread_self(), scheduler, priority_mid);
#endif /* PORTS0_ARCH_PARAGON */
#else
    pthread_setschedattr(pthread_self(), default_thread_attr);
#endif /* PORTS0_ARCH_DCE_THREADS */

    /*
     * Setup thread specific storage which contains
     * a pointer to the _p0_thread_t structure for the thread.
     */
    ports0_thread_key_create(&(ports0_all_global_vars.ports0_thread_t_pointer),
			     NULL);
    _p0_test_rc(rc, "PORTS0: pthread_key_create() failed\n");
    
    ports0_mutex_init(&(thread_mem_mutex), (ports0_mutexattr_t *) NULL);
    next_thread_id = 0;

    /*
     * Initialize the _p0_thread_t structure for this initial thread
     */
    thread = new_thread();
    thread->names = (ports0_thread_name_list_t *) NULL;
    set_tsd(thread);

#ifdef PORTS0_ARCH_DCE_THREADS
    ports0_debug_printf(2, ("_p0_thread_init(): thread priority = %d\n", pthread_getprio(pthread_self())));
#endif

#if PORTS0_DEBUG
	hasThreads = 1;
#endif    
} /* _p0_thread_init() */


/*
 * _p0_thread_shutdown()
 */
void _p0_thread_shutdown()
{
} /* _p0_thread_shutdown() */


/*
 * _p0_thread_abort()
 */
void _p0_thread_abort(int return_code)
{
} /* _p0_thread_abort() */


/*
 * _p0_report_bad_rc()
 */
void _p0_report_bad_rc(int rc, char *message)
{
    char achMessHead[] = "[Thread System]";
    char achDesc[MAX_ERR_SIZE];
    
    if (rc != _P0_THR_SUCCESS)
    {
	ports0_stdio_lock();
	fprintf( stderr, message );
	ports0_stdio_unlock();
	switch( errno )
	{
	case EAGAIN:
	    strcpy(achDesc, "system out of resources (EAGAIN)");
	    break;
	case ENOMEM:
	    strcpy(achDesc, "insufficient memory (ENOMEM)");
	    break;
	case EINVAL:
	    strcpy(achDesc,
		   "invalid value passed to pthread interface (EINVAL)");
	    break;
	case EPERM:
	    strcpy(achDesc,
		   "user does not have adequate permission (EPERM)");
	    break;
	case ERANGE:
	    strcpy(achDesc, "a parameter has an invalid value (ERANGE)");
	    break;
	case EBUSY:
	    strcpy(achDesc, "mutex is locked (EBUSY)");
	    break;
	case EDEADLK:
	    strcpy(achDesc, "deadlock detected (EDEADLK)");
	    break;
	case ESRCH:
	    strcpy(achDesc, "could not find specified thread (ESRCH)");
	    break;
	default:
	    ports0_fatal("%s %s\n%s unknown error number: %d\n",
			achMessHead, message, achMessHead, errno );
	    break;
	}
	/*
	_p0_imprison_thread( "Error code detected\n" );
	*/
	ports0_fatal("%s %s\n%s %s",
		    achMessHead, message, achMessHead, achDesc);
    }
} /* _p0_report_bad_rc() */


/*
 * new_thread()
 *
 * Allocate and return a _p0_thread_t thread structure.
 */
static _p0_thread_t *new_thread( void )
{
    int i;
    _p0_thread_t *new_thread;
    int mem_req_size;
    
    ports0_mutex_lock(&thread_mem_mutex);
    
    if (thread_freelist == NULL)
    {
	mem_req_size = sizeof(_p0_thread_t) * _P0_THREAD_GRAN;
	Ports0Malloc(new_thread(),
		    thread_freelist,
		    _p0_thread_t *, 
		    mem_req_size);
	
	for( i = 0; i < _P0_THREAD_GRAN-1; i++ )
	{
	    thread_freelist[i].next_free = &thread_freelist[i+1];
	}
	thread_freelist[_P0_THREAD_GRAN-1].next_free = NULL;
    }
    new_thread = thread_freelist;
    if (thread_freelist != NULL)
    {
	thread_freelist = thread_freelist->next_free;
    }

    new_thread->id = next_thread_id++;
    
    ports0_mutex_unlock(&thread_mem_mutex);
    
    return (new_thread);
} /* new_thread() */


/*
 * terminate_thread()
 */
static void terminate_thread(_p0_thread_t *thread, void *status)
{
#ifdef PORTS0_PROFILE
/*
    _p0_pablo_log_thread_destruction(node_id, context_id, thread->id);
*/
#endif /* PORTS0_PROFILE */    

    /* Free up the thread storage */
    ports0_mutex_lock(&thread_mem_mutex);
    thread->next_free = thread_freelist;
    thread_freelist = thread;
    ports0_mutex_unlock(&thread_mem_mutex);

    /* Exit the thread */
    pthread_exit(NULL);

} /* terminate_thread() */


/*
 * ports0_thread_exit()
 */
void ports0_thread_exit(void *status)
{
    _p0_thread_t *victim;
    _p0_thread_self(&victim);
    terminate_thread(victim, status);
} /* ports0_thread_exit() */


/*
 * set_tsd()
 *
 * Save the _p0_thread_t thread structure in the pthread's thread
 * specific storage.
 */
static void set_tsd( _p0_thread_t *thread )
{
    ports0_thread_setspecific(ports0_all_global_vars.ports0_thread_t_pointer,
			      (any_t) thread );
} /* set_tsd() */


/*
 * thread_starter()
 *
 * Wrapper to get a Ports0 thread function started.
 */
static void *thread_starter( void *temparg )
{
    _p0_thread_t *thread;
    void *status;

    thread = (_p0_thread_t *) temparg;

    set_tsd(thread);

#ifdef PORTS0_ARCH_DCE_THREADS
    ports0_debug_printf(2, ("thread_starter(): thread priority = %d\n", pthread_getprio(pthread_self())));
#endif
    
#ifdef PORTS0_PROFILE
/*
    _p0_pablo_log_thread_creation(node_id, context_id, thread->id);
*/
#endif /* PORTS0_PROFILE */    
    
    /* Call the user function */
    status = (*thread->user_func)(thread->user_arg);
    
    /* Terminate the thread */
    terminate_thread(thread, status);
  
    return (NULL);
} /* thread_starter() */


/*
 * ports0_thread_create
 */
int ports0_thread_create(ports0_thread_t *user_thread,
			ports0_threadattr_t *attr,
			ports0_thread_func_t func,
			void *user_arg )
{
    int rc;
    _p0_thread_t *thread;
    pthread_t thread_id;

    thread = new_thread();
    thread->names = (ports0_thread_name_list_t *) NULL;
  
    /* Initialize the thread data that needs to be passed to the new thread */
    thread->user_func = func;
    thread->user_arg = user_arg;
  
    rc = pthread_create(&thread_id,
#ifdef PORTS0_ARCH_DCE_THREADS			
			default_thread_attr,
#else			
			&default_thread_attr,
#endif
			thread_starter,
			thread);
    _p0_test_rc( rc, "PORTS0: pthread_create() failed\n" );
    rc = pthread_detach(&thread_id);
    _p0_test_rc( rc, "PORTS0: pthread_detach() failed\n" );

    if (user_thread)
    {
	*user_thread = thread_id;
    }

    return (0);
} /* ports0_thread_create() */


/*
 * ports0_idle_callback()
 *
 * If we have preemptive threads:
 *   Create a normal priority thread that will sit in a
 *   loop calling ports0_poll().
 *
 * If we have non-preemptive threads:
 *   Create a low priority thread which will only become active when
 *   all other threads have suspended.  This thread should:
 *	1) Call _p0_blocking_poll(), to check for available messages.
 *	2) Keep track of idle time.
 *
 *   If a thread module does not support priorities, this same
 *   functionality can be hacked into ports0_cond_wait() using a
 *   counter of suspend threads.
 *
 * TODO: We can probably just use ports0_thread_create() once attributes are
 * implemented.
 */
void ports0_idle_callback(void (*idle_func)())
{
    int rc;
    _p0_thread_t *thread;
    pthread_attr_t my_attr;
    int priority;

    ports0_debug_printf(2, ("ports0_idle_callback(): Creating handler thread\n"));
    
    using_handler_thread = PORTS0_TRUE;
    idle_func_save = idle_func;

    if (preemptive_threads)
    {
	priority = priority_mid;
    }
    else
    {
	priority = priority_min;
    }
    
    thread = new_thread();
    thread->names = (ports0_thread_name_list_t *) NULL;
    thread->user_func = idle_thread_func;
    thread->user_arg = NULL;
    
#ifdef PORTS0_ARCH_DCE_THREADS
    rc = pthread_attr_create(&my_attr);
    _p0_test_rc( rc, "PORTS0: pthread_attr_create() failed\n" );
#else
    rc = pthread_attr_init(&my_attr);
    _p0_test_rc( rc, "PORTS0: pthread_attr_init() failed\n" );
#endif /* PORTS0_ARCH_DCE_THREADS */

#ifndef PORTS0_ARCH_PARAGON
    rc = pthread_attr_setinheritsched(&my_attr, PTHREAD_DEFAULT_SCHED);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setinheritsched() failed\n" );
    rc = pthread_attr_setsched(&my_attr, scheduler);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setsched() failed\n" );
    rc = pthread_attr_setprio(&my_attr, priority);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setprio() failed\n" );
#endif /* PORTS0_ARCH_PARAGON */

#ifdef PORTS0_HACK_STACK_SIZE
    rc = pthread_attr_setstacksize(&my_attr, stack_size);
    _p0_test_rc( rc, "PORTS0: pthread_attr_setstacksize() failed\n" );
#endif
    
    rc = pthread_create(&handler_thread,
#ifdef PORTS0_ARCH_DCE_THREADS			
			my_attr,
#else			
			&my_attr,
#endif
			thread_starter,
			thread);
    _p0_test_rc( rc, "PORTS0: pthread_create() failed\n" );
    
#ifdef PORTS0_ARCH_DCE_THREADS
    rc = pthread_attr_delete(&my_attr);
    _p0_test_rc( rc, "PORTS0: pthread_attr_delete() failed\n" );
#else
    rc = pthread_attr_destroy(&my_attr);
    _p0_test_rc( rc, "PORTS0: pthread_attr_destroy() failed\n" );
#endif /* PORTS0_ARCH_DCE_THREADS */

} /* ports0_idle_callback() */


/*
 * idle_thread_func()
 *
 * TODO: Should count 'unblocked_threads' as they call and return
 * from cond_wait.  Thus, when all threads except this one are blocked,
 * we know we can call _p0_blocking_poll() instead of ports0_poll when
 * using preemptive threads.
 */
static void *idle_thread_func(void *arg)
{
    ports0_debug_printf(2, ("idle_thread_func(): Entering\n"));

    while (!handler_thread_done)
    {
        (*idle_func_save)();
	ports0_debug_printf(6, ("idle_thread_func(): calling yield\n"));
	ports0_thread_yield();
	ports0_debug_printf(6, ("idle_thread_func(): Has been scheduled\n"));
    }

    ports0_debug_printf(2,("idle_thread_func(): Exiting\n"));
    
    return (NULL);
} /* idle_thread_func() */


/*
 * _p0_thread_shutdown_handler_thread()
 */
void _p0_thread_shutdown_handler_thread()
{
    int rc;
    void *status;

    if (using_handler_thread)
    {
	handler_thread_done = PORTS0_TRUE;

	/*
	 * If this is not the idle thread, then wait for
	 * the idle thread to terminate.
	 */
	if (!pthread_equal(pthread_self(), handler_thread))
	{
	    rc = pthread_join( handler_thread, &status );
	    _p0_test_rc( rc, "PORTS0: pthread_join() failed\n" );
	}
    }
} /* _p0_thread_shutdown_handler_thread() */


/*
 * ports0_preemptive_threads
 *
 * Return PORTS0_TRUE (non-zero) if we are using preemptive threads.
 */
bool_t ports0_preemptive_threads()
{
    return (preemptive_threads);
} /* ports0_preemptive_threads() */



/*
 * ports0_thread_key_create()
 */
#undef ports0_thread_key_create
int ports0_thread_key_create(ports0_thread_key_t *key,
			     void (*destructor_func)(void *func))
{
    int rc;
    rc = ports0_macro_thread_key_create(key, destructor_func);
    _p0_test_rc(rc, "PORTS0: ports0_thread_key_create() failed\n");\
    return(rc);
} /* ports0_thread_key_create() */


/*
 * ports0_thread_key_delete()
 */
#undef ports0_thread_key_delete
int ports0_thread_key_delete(ports0_thread_key_t *key)
{
    int rc;
    rc = ports0_macro_thread_key_delete(key);
    _p0_test_rc(rc, "PORTS0: ports0_thread_key_delete() failed\n");\
    return(rc);
} /* ports0_thread_key_delete() */


/*
 * ports0_thread_setspecific()
 */
#undef ports0_thread_setspecific
int ports0_thread_setspecific(ports0_thread_key_t key,
			      void *value)
{
    int rc;
    rc = ports0_macro_thread_setspecific(key, value);
    _p0_test_rc(rc, "PORTS0: ports0_thread_setspecific() failed\n");
    return(rc);
} /* ports0_thread_setspecific() */


/*
 * ports0_thread_getspecific()
 */
#undef ports0_thread_getspecific
int ports0_thread_getspecific(ports0_thread_key_t key,
			      void **value)
{
    int rc;
    rc = ports0_macro_thread_getspecific(key, value);
    _p0_test_rc(rc, "PORTS0: ports0_thread_getspecific() failed\n");
    return(rc);
} /* ports0_thread_getspecific() */


/*
 * ports0_thread_self()
 */
#undef ports0_thread_self
ports0_thread_t ports0_thread_self(void)
{
  return(ports0_macro_thread_self());
} /* ports0_thread_self() */


/*
 * ports0_thread_equal()
 */
#undef ports0_thread_equal
int ports0_thread_equal(ports0_thread_t t1,
			ports0_thread_t t2)
{
    return (ports0_macro_thread_equal(t1, t2));
} /* ports0_thread_equal() */


/*
 * ports0_thread_once()
 */
#undef ports0_thread_once
int ports0_thread_once(ports0_thread_once_t *once_control,
		       void (*init_routine)())
{
    return (ports0_macro_thread_once(once_control, init_routine));
} /* ports0_thread_once() */


/*
 * ports0_thread_yield
 */
#undef ports0_thread_yield
void ports0_thread_yield(void)
{
    ports0_macro_thread_yield();
} /* ports0_thread_yield() */


/*
 * ports0_mutex_init()
 */
#undef ports0_mutex_init
int ports0_mutex_init(ports0_mutex_t *mut, ports0_mutexattr_t *attr)
{
    int rc;
    _P0_INIT_START_MAGIC_COOKIE(mut);
    rc = ports0_macro_mutex_init(mut, attr);
    _P0_INIT_END_MAGIC_COOKIE(mut);
    _p0_test_rc( rc, "PORTS0: pthread_mutex_init() failed\n" );
    return(rc);
} /* ports0_mutex_init() */


/*
 *  ports0_mutex_destroy()
 */
#undef ports0_mutex_destroy
int ports0_mutex_destroy(ports0_mutex_t *mut)
{
    int rc; 
    PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "destroy" ); 
    rc = ports0_macro_mutex_destroy(mut);
    _p0_test_rc( rc, "PORTS0: pthread_mutex_destroy() failed\n" );
    return(rc);
} /* ports0_mutex_destroy() */


/* 
 *  ports0_mutex_lock()
 */
#undef ports0_mutex_lock
int ports0_mutex_lock(ports0_mutex_t *mut)
{
    int rc;
    PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "lock before" );
    rc = ports0_macro_mutex_lock(mut);
    _p0_test_rc( rc, "PORTS0: pthread_mutex_lock() failed\n" );
    PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "lock after" );
    return(rc);
} /* ports0_mutex_lock() */


/* 
 *  ports0_mutex_trylock()
 */
#undef ports0_mutex_trylock
int ports0_mutex_trylock(ports0_mutex_t *mut)
{
    int rc;
    PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "trylock before" );
    rc = ports0_macro_mutex_trylock(mut);
    if (rc != EBUSY)
    {
	_p0_test_rc( rc, "PORTS0: pthread_mutex_trylock() failed\n" );
    }
    PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "trylock after" );
    return(rc);
} /* ports0_mutex_trylock() */


/*
 *  ports0_mutex_unlock()
 */
#undef ports0_mutex_unlock
int ports0_mutex_unlock(ports0_mutex_t *mut)
{
    int rc;
    PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "unlock before" );
    rc = ports0_macro_mutex_unlock(mut);
    _p0_test_rc( rc, "PORTS0: pthread_mutex_unlock() failed\n" );
    PORTS0_INTERROGATE( mut, _P0_MUTEX_T, "unlock after" );
    return(rc);
} /* ports0_mutex_unlock() */


/*
 * ports0_cond_init()
 */
#undef ports0_cond_init
int ports0_cond_init(ports0_cond_t *cv, ports0_condattr_t *attr)
{
    int rc;
    _P0_INIT_START_MAGIC_COOKIE(cv); 
    rc = ports0_macro_cond_init(cv, attr);
    _p0_test_rc( rc, "PORTS0: pthread_cond_init() failed\n" );
    _P0_INIT_END_MAGIC_COOKIE(cv); 
    return(rc);
} /* ports0_cond_init() */


/*
 *  ports0_cond_destroy()
 */
#undef ports0_cond_destroy
int ports0_cond_destroy(ports0_cond_t *cv)
{
    int rc; 
    PORTS0_INTERROGATE( cv, _P0_COND_T, "destroy" ); 
    rc = ports0_macro_cond_destroy(cv);
    _p0_test_rc( rc, "PORTS0: pthread_cond_destroy() failed\n" );
    return(rc);
} /* ports0_cond_destroy() */


/*
 *  ports0_cond_wait()
 */
#undef ports0_cond_wait
int ports0_cond_wait(ports0_cond_t *cv, ports0_mutex_t *mut)
{
    int rc; \
    PORTS0_INTERROGATE( cv, _P0_COND_T, "wait before" ); \
    rc = ports0_macro_cond_wait(cv, mut);
    PORTS0_INTERROGATE( cv, _P0_COND_T, "wait after" ); \
    _p0_test_rc( rc, "PORTS0: pthread_cond_wait() failed\n" );
    return(rc);
} /* ports0_cond_wait() */


/*
 *  ports0_cond_signal()
 */
#undef ports0_cond_signal
int ports0_cond_signal(ports0_cond_t *cv)
{
    int rc; 
    PORTS0_INTERROGATE( cv, _P0_COND_T, "signal" ); 
    rc = ports0_macro_cond_signal(cv); 
    _p0_test_rc( rc, "PORTS0: pthread_cond_signal() failed\n" );
    return(rc);
} /* ports0_cond_signal () */


/*
 *  ports0_cond_broadcast()
 */
#undef ports0_cond_broadcast
int ports0_cond_broadcast(ports0_cond_t *cv)
{
    int rc; 
    PORTS0_INTERROGATE( cv, _P0_COND_T, "broadcast" ); 
    rc = ports0_macro_cond_broadcast(cv); 
    _p0_test_rc( rc, "PORTS0: pthread_cond_broadcast() failed\n" );
    return(rc);
} /* ports0_cond_broadcast() */
